package ca.recrown.islandsurvivalcraft.caching;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.ReentrantLock;

public class Cache<Key, Value> {
    private final int maxSize;
    private final ConcurrentHashMap<Key, CacheValue<Value>> data;
    private final ConcurrentLinkedQueue<Key> occurrenceOrder;
    private ReentrantLock cleaningLock;

    public Cache(int maxSize) {
        data = new ConcurrentHashMap<>(maxSize);
        occurrenceOrder = new ConcurrentLinkedQueue<>();
        this.maxSize = maxSize;
        cleaningLock = new ReentrantLock(true);
    }

    public Cache() {
        this(1024);
    }

    public void setValue(Key key, Value value) {
        CacheValue<Value> previous = data.get(key);
        CacheValue<Value> fresh = new CacheValue<>(value);
        if (previous != null) fresh.occurrence += previous.occurrence;
        occurrenceOrder.add(key);
        data.put(key, fresh);

        if (data.size() > maxSize) {
            int occ = 0;
            do {
                cleaningLock.lock();
                try {
                    Key potentialKey = occurrenceOrder.poll();
                    CacheValue<Value> potential = data.get(potentialKey);
                    potential.occurrence--;
                    occ = potential.occurrence;
                    if (occ < 1) {
                        data.remove(potentialKey);
                    } 
                } finally {
                    cleaningLock.unlock();
                }
            } while (occ > 0);
        }
    }

    /**
     * Retrieves cached value associated to the key.
     * Uses equals.
     * If does not exist, returns null.
     * @param key the key associated with the value.
     * @return the value associated to the key.
     */
    public Value getValue(Key key) {
        CacheValue<Value> res = null;
        cleaningLock.lock();
        try {
            res = data.get(key);
            if (res == null) return null;
            res.occurrence++;
            occurrenceOrder.add(key);
        } finally {
            cleaningLock.unlock();
        }

        return res.data;
    }

    public void clearCache() {
        data.clear();
        occurrenceOrder.clear();
    }
}